import numpy

from ..io.save_xas import save_xdi_data
from ..io.types import XdiData


def test_save_xas(tmp_path):
    xdi_filename = tmp_path / "example.xdi"

    xdi_metadata = {
        "Beamline": {},
        "Scan": {
            "start_time": "2025-09-02T17:26:14.195421+02:00",
            "end_time": "2025-09-02T17:44:38.146967+02:00",
        },
        "Facility": {"current": 84.5, "fillingMode": "16 bunch"},
        "Mono": {"name": "Si 311", "d_spacing": 1.637039},
    }
    data = {
        "column_names": ["A", "B"],
        "column_data": [numpy.asarray([1, 2, 3]), numpy.asarray([4, 5, 6])],
        "xdi_metadata": xdi_metadata,
    }
    xdi_data = XdiData(**data)

    save_xdi_data(xdi_data, xdi_filename)

    with open(xdi_filename, "r") as f:
        content = f.read()

    expected = """# XDI/1.0 GSE/1.0
# Data collected at BM08-LISA of ESRF preproccessed with the beamline Calibration Tool
# Facility.name: ESRF
# Facility.xray_source: bending magnet
# Facility.current: 84.5
# Facility.fillingMode: 16 bunch
# Beamline.name: BM08-LISA
# Scan.start_time: 2025-09-02T17:26:14.195421+02:00
# Scan.end_time: 2025-09-02T17:44:38.146967+02:00
# Mono.name: Si 311
# Mono.d_spacing: 1.637039
# Mono.notes: LNT cooled
# Column.0: A
# Column.1: B
# -------------
# A B
1.0000000 4.0000000
2.0000000 5.0000000
3.0000000 6.0000000
"""

    print(content)

    assert content == expected



import os

from ewokscore.model import BaseInputModel
from ewokscore.model import BaseOutputModel
from ewokscore.task import Task
from pydantic import Field

from ..io.save_xas import save_xdi_data
from ..io.types import XdiData


class InputModel(BaseInputModel):
    filename: str = Field(..., description="XDI file name.", examples=[])
    xdi_data: XdiData = Field(..., description="XDI data to be saved.", examples=[])


class OutputModel(BaseOutputModel):
    output_filename: str = Field(..., description="XDI file name.")


class SaveXasXdi(Task, input_model=InputModel, output_model=OutputModel):
    """Save one XAS scan as an XDI file."""

    def run(self):
        save_xdi_data(self.inputs.xdi_data, self.inputs.filename)
        self.outputs.output_filename = os.path.abspath(self.inputs.filename)



import pytest
from ewokscore.bindings import execute_graph


@pytest.mark.parametrize("mono_counter", ["mono_enc", "energy"])
def test_convert(example_hdf5_path, mono_counter, tmp_path):
    output_filename = tmp_path / "example_scan13.xdi"
    entry_name = "13.1"
    optional_counters = [
        "I0_eh1",
        "I1_eh1",
        "IX_eh1",
        "I0_eh2",
        "I1_eh2",
        "IX_eh2",
        "IR_eh2",
        "volt1",
        "volt2",
    ]
    optional_mca_counters = ["Te_Ka"]
    livetime_normalization = 3.0
    crystal_motor = "c_sel"

    load_identifier = "ReadXasHdf5"
    save_identifier = "SaveXasXdi"
    inputs = [
        {
            "task_identifier": load_identifier,
            "name": "filename",
            "value": str(example_hdf5_path),
        },
        {
            "task_identifier": load_identifier,
            "name": "entry_name",
            "value": entry_name,
        },
        {
            "task_identifier": load_identifier,
            "name": "mono_counter",
            "value": mono_counter,
        },
        {
            "task_identifier": load_identifier,
            "name": "crystal_motor",
            "value": crystal_motor,
        },
        {
            "task_identifier": load_identifier,
            "name": "optional_counters",
            "value": optional_counters,
        },
        {
            "task_identifier": load_identifier,
            "name": "optional_mca_counters",
            "value": optional_mca_counters,
        },
        {
            "task_identifier": load_identifier,
            "name": "livetime_normalization",
            "value": livetime_normalization,
        },
        {
            "task_identifier": save_identifier,
            "name": "filename",
            "value": str(output_filename),
        },
    ]

    result = execute_graph(_WORKFLOW, inputs=inputs, outputs=[{"all": False}])
    xdi_file = result["output_filename"]
    assert xdi_file == str(output_filename)


_WORKFLOW = {
    "graph": {"id": "convert", "schema_version": "1.1"},
    "nodes": [
        {
            "id": "load",
            "default_inputs": [],
            "task_type": "class",
            "task_identifier": "ewoksbm08.tasks.read_xas.ReadXasHdf5",
        },
        {
            "id": "save",
            "default_inputs": [],
            "task_type": "class",
            "task_identifier": "ewoksbm08.tasks.save_xas.SaveXasXdi",
        },
    ],
    "links": [
        {
            "source": "load",
            "target": "save",
            "data_mapping": [{"source_output": "xdi_data", "target_input": "xdi_data"}],
        }
    ],
}



from blissdata.h5api import dynamic_hdf5
from ewokscore.model import BaseInputModel
from ewokscore.model import BaseOutputModel
from ewokscore.task import Task
from pydantic import Field

from ..io.read_xas import read_xas_hdf5
from ..io.types import XdiData


class InputModel(BaseInputModel):
    filename: str = Field(..., description="Bliss HDF5 file name.")
    entry_name: str = Field(
        ..., description="NeXus entry name.", examples=["1.1", "5.2"]
    )
    mono_counter: str = Field(
        ..., description="Counter of the energy or theta value of the monochromator."
    )
    crystal_motor: str = Field(
        ..., description="Motor that selects the monochromator crystal."
    )
    optional_counters: list[str | None] = Field(
        None, description="Counters other than the energy to be added to the XDI data."
    )
    optional_mca_counters: list[str | None] = Field(
        None,
        description="MCA counters (ROI's) other than the energy to be added to the XDI data.",
    )
    livetime_normalization: float | None = Field(
        None,
        description="Live-time normalization in seconds for 'optional_mca_counters' "
        "(`None`: no normalization, `<=0` the median of the elapsed time).",
    )
    skip_mca: dict[str, list[int | None]] = Field(
        None,
        description="Skip detectors from MCA controllers.",
        examples=[{"d08xmap": [1, 3]}],
    )
    retry_timeout: float | None = Field(
        0.0, description="Timeout when trying to read data from HDF5."
    )
    mono_edge_theoretical: float | None = Field(
        None,
        description="The theoretical edge position in 'mono_counter' units.",
    )
    mono_edge_experimental: float | None = Field(
        None,
        description="The experimental edge position in 'mono_counter' units.",
    )


class OutputModel(BaseOutputModel):
    xdi_data: XdiData = Field(..., description="")


class ReadXasHdf5(Task, input_model=InputModel, output_model=OutputModel):
    """Read one XAS scan with XDI structured metadata."""

    def run(self):
        with dynamic_hdf5.File(
            self.inputs.filename, retry_timeout=self.inputs.retry_timeout
        ) as h5file:
            h5scan = h5file[self.inputs.entry_name]
            self.outputs.xdi_data = read_xas_hdf5(
                h5scan,
                mono_counter=self.inputs.mono_counter,
                crystal_motor=self.inputs.crystal_motor,
                counters=self.inputs.optional_counters,
                mca_counters=self.inputs.optional_mca_counters,
                livetime_normalization=self.inputs.livetime_normalization,
                skip_mca=self.inputs.skip_mca,
                mono_edge_theoretical=self.inputs.mono_edge_theoretical,
                mono_edge_experimental=self.inputs.mono_edge_experimental,
            )



import numpy
from pydantic import BaseModel


class XdiBaseModel(BaseModel):
    def header_lines(self, prefix: str = None) -> list[str]:
        lines = []

        for field_name in type(self).model_fields:
            value = getattr(self, field_name)

            if value is None:
                continue

            if isinstance(value, XdiBaseModel):
                lines.extend(
                    value.header_lines(
                        prefix=f"{prefix}.{field_name}" if prefix else field_name
                    )
                )
            else:
                if prefix:
                    lines.append(f"{prefix}.{field_name}: {value}")
                else:
                    lines.append(str(value))

        return lines


class XdiFacility(XdiBaseModel):
    name: str = "ESRF"
    xray_source: str = "bending magnet"
    current: float | None = None
    fillingMode: str | None = None


class XdiBeamline(XdiBaseModel):
    name: str = "BM08-LISA"


class XdiScan(XdiBaseModel):
    start_time: str | None = None
    end_time: str | None = None


class XdiMono(XdiBaseModel):
    name: str | None = None
    d_spacing: float | None = None
    notes: str | None = "LNT cooled"


class XdiSample(XdiBaseModel):
    name: str | None = None


class XdiMetadata(XdiBaseModel):
    """XDI SPECS:

    https://github.com/XraySpectroscopy/XAS-Data-Interchange/blob/master/specification/spec.md
    """

    Scheme: str = "XDI/1.0 GSE/1.0"
    Title: str = (
        "Data collected at BM08-LISA of ESRF preproccessed with the beamline Calibration Tool"
    )
    Facility: XdiFacility
    Beamline: XdiBeamline
    Scan: XdiScan | None = None
    Mono: XdiMono | None = None
    Sample: XdiSample | None = None


class XdiData(BaseModel, arbitrary_types_allowed=True):
    column_names: list[str]
    column_data: list[numpy.ndarray]
    xdi_metadata: XdiMetadata

    def header_lines(self) -> list[str]:
        lines = self.xdi_metadata.header_lines()
        if self.column_names:
            for i, name in enumerate(self.column_names):
                lines.append(f"Column.{i}: {name}")
            lines.append("-------------")
            lines.append(" ".join(self.column_names))
        return lines



import importlib.resources as importlib_resources
from contextlib import contextmanager
from pathlib import Path
from typing import Generator


@contextmanager
def resource_path(*args) -> Generator[Path, None, None]:
    """
    .. code-block:: python

        with resource_path("example.h5") a path:
            ...
    """
    source = importlib_resources.files(__name__).joinpath(*args)
    with importlib_resources.as_file(source) as path:
        if not path.is_file():
            raise FileNotFoundError(f"Not a file: '{path}'")
        yield path



import logging
import re

import h5py
import numpy
from blissdata.h5api.dynamic_hdf5 import Group

from .types import XdiData

logger = logging.getLogger(__name__)

GroupType = Group | h5py.Group


def read_xas_hdf5(
    h5scan: GroupType,
    mono_counter: str,
    crystal_motor: str,
    counters: list[str],
    mca_counters: list[str],
    livetime_normalization: float | None = None,
    skip_mca: dict[str, list[int | None]] = None,
    mono_edge_theoretical: float | None = None,
    mono_edge_experimental: float | None = None,
) -> XdiData:
    """
    :param h5scan: NXentry group of one Bliss scan.
    :param mono_counter: energy or theta counter.
    :param crystal_motor: defines the mono crystal that is selected.
    :param counters: map name in the measurement group to name in XDI.
    :param mca_counters: counters the be extracted for all MCA detector. Use `skip_mca` to skip some.
    :param livetime_normalization: normalize `mca_counters` to a common live-time.
    :param skip_mca: scan MCA detector.
    :param mono_edge_theoretical: theoretical edge position in `mono_counter` units.
    :param mono_edge_experimental: experimental edge position in `mono_counter` units.
    :returns: XAS scan object
    """
    xdi_metadata = {
        "Beamline": {},
    }

    # Scan
    start_time = h5scan["start_time"][()].decode()
    end_time = h5scan["end_time"][()].decode()
    xdi_metadata["Scan"] = {"start_time": start_time, "end_time": end_time}

    # Facility
    instrument = h5scan["instrument"]
    current = instrument["machine/current"][()]
    filling_mode = instrument["machine/filling_mode"][()].decode()
    xdi_metadata["Facility"] = {
        "current": current,
        "fillingMode": filling_mode,
    }

    # Mono
    si_hkl = _get_si_hkl(h5scan, crystal_motor)
    d_spacing = _mono_d_spacing(*si_hkl)
    si_hkl_string = "".join(map(str, si_hkl))
    xdi_metadata["Mono"] = {"name": f"Si {si_hkl_string}", "d_spacing": d_spacing}

    # Energy in eV from monochromator counter
    measurement = h5scan["measurement"]
    if mono_counter not in measurement:
        raise ValueError(
            f"{mono_counter!r} is not in {measurement.file.filename}::{measurement.name}"
        )
    mono_dset = measurement[mono_counter]
    energy = _convert_to_energy(
        mono_dset, d_spacing, mono_edge_theoretical, mono_edge_experimental
    )
    column_names = ["energy"]
    column_data = [energy]

    # Regular counters
    if counters:
        ctr_column_names, ctr_column_data = _get_counter_columns(h5scan, counters)
        column_names += ctr_column_names
        column_data += ctr_column_data

    # MCA counters
    if mca_counters:
        mca_column_names, mca_column_data = _get_mca_columns(
            h5scan, mca_counters, livetime_normalization, skip_mca or {}
        )
        column_names += mca_column_names
        column_data += mca_column_data

    # Ensure all columns have the same length
    lengths = [len(values) for values in column_data]
    if len(set(lengths)) > 1:
        n = min(lengths)
        column_data = [values[:n] for values in column_data]

    return XdiData(
        column_names=column_names, column_data=column_data, xdi_metadata=xdi_metadata
    )


def _convert_to_energy(
    mono_dset: h5py.Dataset,
    d_spacing: float,
    mono_edge_theoretical: float | None,
    mono_edge_experimental: float | None,
) -> numpy.ndarray:
    """Energy in eV from theta in radians or energy in keV or eV."""
    mono_values = mono_dset[()]
    if mono_edge_theoretical is not None and mono_edge_experimental is not None:
        mono_values += round(mono_edge_theoretical - mono_edge_experimental, 7)

    # TODO: should this be fixed in blissdata? When there are no units, it waits retry_timeout
    # seconds and then returns, even when the scan is over.
    mono_units = dict(mono_dset.attrs).get("units", "").lower()
    if mono_units == "ev":
        return mono_values
    if mono_units == "kev":
        return mono_values * 1000
    if mono_units in ("deg", "degree", "degrees"):
        mono_values = numpy.radians(mono_values)
        mono_units = ""
    if mono_units in ("", "rad", "radian", "radians"):
        return _theta_to_energy(mono_values, d_spacing)
    raise ValueError(f"Unknown monochromator counter units {mono_units!r}")


def _mono_d_spacing(si_h: int, si_k: int, si_l: int) -> float:
    """D-spacing from Si HKL planes in Å.

    Cubic lattice: dhkl(Å) = A0(Å) / sqrt(h^2+k^2+l^2)
    """
    si_lattice_constant = 5.42944537  # Å
    return si_lattice_constant / numpy.sqrt(si_h**2 + si_k**2 + si_l**2)


def _theta_to_energy(theta_angle: float | numpy.ndarray, d_spacing: float) -> float:
    """Converts the monochromator angle in radians into energy
    given a monochromator crystal d-spacing.

    Bragg's Law: E(eV) = (h * c) / (2 * dhkl(Å) * sinθ)
    """
    hc = 12398.42014541  # Å/eV
    sin_theta = numpy.sin(theta_angle)
    return numpy.round(hc / (2 * d_spacing * sin_theta), 3)


def _get_counter_columns(
    h5scan: GroupType, counters: list[str]
) -> tuple[list[str], list[numpy.ndarray]]:
    measurement = h5scan["measurement"]
    column_names = []
    column_data = []
    for name in counters:
        if name in measurement:
            column_names.append(name)
            column_data.append(measurement[name][()])
        else:
            logger.warning(
                "Skip: regular counter %r not in %r",
                name,
                f"{measurement.file.filename}::{measurement.name}",
            )
    return column_names, column_data


def _get_mca_columns(
    h5scan: GroupType,
    mca_counters: list[str],
    livetime_normalization: float | None,
    skip_mca: dict[str, list[int]],
) -> tuple[list[str], list[numpy.ndarray]]:
    instrument = h5scan["instrument"]
    mca_detectors = _get_mca_detectors(instrument)
    if not mca_detectors:
        logger.warning(
            "No MCA detector found in %s::%s", h5scan.file.filename, instrument.name
        )
        return [], []

    if livetime_normalization is not None and livetime_normalization <= 0:
        livetime_normalization = _get_count_time(h5scan) or -1

    column_short_names = []
    column_prefixes = []
    column_data = []

    for (controller_name, detector_index), detector in mca_detectors.items():
        skip_indices = skip_mca.get(controller_name, [])
        if detector_index in skip_indices:
            continue
        for counter_name in mca_counters:
            values = _mca_counter_values(detector, counter_name, livetime_normalization)
            if values is None:
                continue
            column_short_name = f"{counter_name}_{detector_index}"
            column_short_names.append(column_short_name)
            column_prefixes.append(controller_name)
            column_data.append(values)

    if len(column_short_names) != len(set(column_short_names)):
        column_names = [
            f"{prefix}_{name}"
            for prefix, name in zip(column_prefixes, column_short_names)
        ]
    else:
        column_names = column_short_names

    return column_names, column_data


def _get_mca_detectors(instrument: GroupType) -> dict[tuple[str, int], GroupType]:
    """Find all HDF5 groups that belong to a detector index of an MCA controller."""
    groups = {}
    for name in instrument:
        item = instrument[name]
        if not isinstance(item, GroupType):
            continue

        if "type" in item:
            detector_type = item["type"][()].decode()
        else:
            continue

        if detector_type == "mca":
            match = re.match(r"^(.*?)_det(\d+)$", name)
            if match:
                controller_name = match.group(1)
                detector_index = int(match.group(2))
                groups[(controller_name, detector_index)] = item
    return groups


def _mca_counter_values(
    detector: GroupType,
    counter_name: str,
    livetime_normalization: float | None,
) -> numpy.ndarray | None:
    """Return the counter values (if present) for a detector index of an MCA controller."""
    if counter_name not in detector:
        logger.warning(
            "Skip: MCA counter %r not in %r",
            counter_name,
            f"{detector.file.filename}::{detector.name}",
        )
        return None

    values = detector[counter_name][()]

    # Extract live-time to normalize to from elapsed time per point
    if livetime_normalization is not None and livetime_normalization <= 0:
        if "elapsed_time" in detector:
            elapsed_time = detector["elapsed_time"][()]
            livetime_normalization = numpy.nanmedian(elapsed_time)
            logger.info(
                "Median elapsed time per scan point: %s seconds", livetime_normalization
            )
        else:
            livetime_normalization = None
            logger.warning(
                "Skip live-time normalization of MCA counter %r since %r is not in %r",
                counter_name,
                "elapsed_time",
                f"{detector.file.filename}::{detector.name}",
            )

    # Normalize counts to a common live-time
    if livetime_normalization is not None:
        if "live_time" in detector:
            effective_livetime = detector["live_time"][()]
            if values.size != effective_livetime.size:
                n = min(values.size, effective_livetime.size)
                values = values[:n]
                effective_livetime = effective_livetime[:n]
            values = values / effective_livetime * livetime_normalization
        else:
            logger.warning(
                "Skip live-time normalization of MCA counter %r since %r is not in %r",
                counter_name,
                "live_time",
                f"{detector.file.filename}::{detector.name}",
            )

    return values


def _get_count_time(h5scan: GroupType) -> float | None:
    if "scan_parameters" in h5scan:
        scan_parameters = h5scan["scan_parameters"]
        if "count_time" in scan_parameters:
            count_time = scan_parameters["count_time"][()]
            logger.info("User preset count time: %s seconds", count_time)
            return count_time


def _get_si_hkl(h5scan: GroupType, crystal_motor: str) -> tuple[int, int, int]:
    positioners = h5scan["instrument/positioners"]
    if crystal_motor not in positioners:
        raise ValueError(
            f"{crystal_motor!r} is not in {positioners.file.filename}::{positioners.name}"
        )
    if positioners[crystal_motor][()] > 0:
        return 3, 1, 1
    else:
        return 1, 1, 1



import logging
import os

import numpy

from .types import XdiData

logger = logging.getLogger(__name__)


def save_xdi_data(xdi_data: XdiData, filename: str) -> None:
    parent = os.path.dirname(filename)
    if parent:
        os.makedirs(parent, exist_ok=True)

    if os.path.exists(filename):
        logger.warning("Overwrite %r", filename)

    header = "\n".join(xdi_data.header_lines())
    data = numpy.column_stack(xdi_data.column_data)
    numpy.savetxt(
        filename,
        data,
        fmt="%.7f",
        header=header,
    )



import h5py

from ..io.read_xas import read_xas_hdf5


def test_read_xas(example_hdf5_path):
    entry_name = "13.1"
    mono_counter = "mono_enc"
    counters = [
        "I0_eh1",
        "I1_eh1",
        "IX_eh1",
        "I0_eh2",
        "I1_eh2",
        "IX_eh2",
        "IR_eh2",
        "volt1",
        "volt2",
    ]
    mca_counters = ["Te_Ka"]
    crystal_motor = "c_sel"

    with h5py.File(example_hdf5_path, mode="r") as h5file:
        xdi_data = read_xas_hdf5(
            h5file[entry_name], mono_counter, crystal_motor, counters, mca_counters
        )

    expected = [
        "energy",
        "I0_eh1",
        "I1_eh1",
        "IX_eh1",
        "I0_eh2",
        "I1_eh2",
        "IX_eh2",
        "IR_eh2",
        "volt1",
        "volt2",
        "Te_Ka_0",
        "Te_Ka_1",
        "Te_Ka_2",
        "Te_Ka_3",
    ]
    assert xdi_data.column_names == expected
